# SPDX-License-Identifier: BSD-2-Clause
# Copyright (C) 2016 - 2021 Intel Corporation.

import os
from python_framework.cmd_helper import CMD_helper


def _get_libautohbw_path():
    for p in ["/usr/lib64/libautohbw.so.0", "/usr/lib/libautohbw.so.0"]:
        if os.path.isfile(p):
            return p
    raise Exception("Cannot find library path in OS")


class Test_autohbw(object):
    binary = "../autohbw_test_helper"
    fail_msg = "Test failed with:\n {0}"
    test_prefix = "AUTO_HBW_LOG=2 LD_PRELOAD=%s " % _get_libautohbw_path()
    memkind_malloc_log = "In my memkind malloc"
    memkind_calloc_log = "In my memkind calloc"
    memkind_realloc_log = "In my memkind realloc"
    memkind_posix_memalign_log = "In my memkind align"
    memkind_free_log = "In my memkind free"
    cmd_helper = CMD_helper()

    def test_TC_MEMKIND_autohbw_malloc_and_free(self):
        """ This test executes ./autohbw_test_helper with LD_PRELOAD that is
            overriding malloc() and free() to equivalent autohbw functions"""
        command = (self.test_prefix
                   + self.cmd_helper.get_command_path(self.binary) + " malloc")
        print("Executing command: {0}".format(command))
        output, retcode = self.cmd_helper.execute_cmd(command, sudo=False)
        assert retcode == 0, \
            self.fail_msg.format("\nError: autohbw_test_helper",
                                 " returned {0} \noutput: {1}"
                                 .format(retcode, output))
        assert self.memkind_malloc_log in output, self.fail_msg.format(
            "\nError: malloc was not overridden",
            " by autohbw equivalent \noutput: {0}").format(output)
        assert self.memkind_free_log in output, self.fail_msg.format(
            "\nError: free was not overridden",
            " by autohbw equivalent \noutput: {0}").format(output)

    def test_TC_MEMKIND_autohbw_calloc_and_free(self):
        """ This test executes ./autohbw_test_helper with
            LD_PRELOAD that is overriding calloc() and free()
            to equivalent autohbw functions"""
        command = (self.test_prefix
                   + self.cmd_helper.get_command_path(self.binary) + " calloc")
        print("Executing command: {0}".format(command))
        output, retcode = self.cmd_helper.execute_cmd(command, sudo=False)
        assert retcode == 0, \
            self.fail_msg.format("\nError: autohbw_test_helper",
                                 " returned {0} \noutput: {1}"
                                 .format(retcode, output))
        assert self.memkind_calloc_log in output, self.fail_msg.format(
            "\nError: calloc was not overridden by",
            " autohbw equivalent \noutput: {0}").format(output)
        assert self.memkind_free_log in output, self.fail_msg.format(
            "Error: free was not overridden by",
            " autohbw equivalent \noutput: {0}").format(output)

    def test_TC_MEMKIND_autohbw_realloc_and_free(self):
        """ This test executes ./autohbw_test_helper with LD_PRELOAD that is
            overriding realloc() and free() to equivalent autohbw functions"""
        command = (self.test_prefix
                   + self.cmd_helper.get_command_path(self.binary)
                   + " realloc")
        print("Executing command: {0}".format(command))
        output, retcode = self.cmd_helper.execute_cmd(command, sudo=False)
        assert retcode == 0, self.fail_msg.format(
            "\nError: autohbw_test_helper returned {0} \noutput: {1}"
            .format(retcode, output))
        assert self.memkind_realloc_log in output, self.fail_msg.format(
            "\nError: realloc was not overridden by",
            " autohbw equivalent \noutput: {0}").format(output)
        assert self.memkind_free_log in output, self.fail_msg.format(
            "\nError: free was not overridden by",
            " autohbw equivalent \noutput: {0}").format(output)

    def test_TC_MEMKIND_autohbw_posix_memalign_and_free(self):
        """ This test executes ./autohbw_test_helper with LD_PRELOAD
            that is overriding posix_memalign() and free()
            to equivalent autohbw functions"""
        command = (self.test_prefix
                   + self.cmd_helper.get_command_path(self.binary)
                   + " posix_memalign")
        print("Executing command: {0}".format(command))
        output, retcode = self.cmd_helper.execute_cmd(command, sudo=False)
        assert retcode == 0, self.fail_msg.format(
            "\nError: autohbw_test_helper returned {0} \noutput: {1}"
            .format(retcode, output))
        assert self.memkind_posix_memalign_log in output, self.fail_msg.format(
            "\nError: posix_memalign was not overridden by",
            " autohbw equivalent \noutput: {0}").format(output)
        assert self.memkind_free_log in output, self.fail_msg.format(
            "\nError: free was not overridden by",
            " autohbw equivalent \noutput: {0}").format(output)



#!/usr/bin/env python

import sys
from itertools import combinations
from os import uname
from multiprocessing import cpu_count
from subprocess import call

# Later, we want to test extended vaddr support.  Apparently, the "real" way of
# checking this is flaky on OS X.
bits_64 = sys.maxsize > 2**32

nparallel = cpu_count() * 2

uname = uname()[0]

if "BSD" in uname:
    make_cmd = 'gmake'
else:
    make_cmd = 'make'

def powerset(items):
    result = []
    for i in xrange(len(items) + 1):
        result += combinations(items, i)
    return result

possible_compilers = []
for cc, cxx in (['gcc', 'g++'], ['clang', 'clang++']):
    try:
        cmd_ret = call([cc, "-v"])
        if cmd_ret == 0:
            possible_compilers.append((cc, cxx))
    except:
        pass
possible_compiler_opts = [
    '-m32',
]
possible_config_opts = [
    '--enable-debug',
    '--enable-prof',
    '--disable-stats',
    '--enable-opt-safety-checks',
]
if bits_64:
    possible_config_opts.append('--with-lg-vaddr=56')

possible_malloc_conf_opts = [
    'tcache:false',
    'dss:primary',
    'percpu_arena:percpu',
    'background_thread:true',
]

print 'set -e'
print 'if [ -f Makefile ] ; then %(make_cmd)s relclean ; fi' % {'make_cmd': make_cmd}
print 'autoconf'
print 'rm -rf run_tests.out'
print 'mkdir run_tests.out'
print 'cd run_tests.out'

ind = 0
for cc, cxx in possible_compilers:
    for compiler_opts in powerset(possible_compiler_opts):
        for config_opts in powerset(possible_config_opts):
            for malloc_conf_opts in powerset(possible_malloc_conf_opts):
                if cc is 'clang' \
                  and '-m32' in possible_compiler_opts \
                  and '--enable-prof' in config_opts:
                    continue
                config_line = (
                    'EXTRA_CFLAGS=-Werror EXTRA_CXXFLAGS=-Werror '
                    + 'CC="{} {}" '.format(cc, " ".join(compiler_opts))
                    + 'CXX="{} {}" '.format(cxx, " ".join(compiler_opts))
                    + '../../configure '
                    + " ".join(config_opts) + (' --with-malloc-conf=' +
                    ",".join(malloc_conf_opts) if len(malloc_conf_opts) > 0
                    else '')
                )

                # We don't want to test large vaddr spaces in 32-bit mode.
		if ('-m32' in compiler_opts and '--with-lg-vaddr=56' in
                  config_opts):
		    continue

                # Per CPU arenas are only supported on Linux.
                linux_supported = ('percpu_arena:percpu' in malloc_conf_opts \
                  or 'background_thread:true' in malloc_conf_opts)
                # Heap profiling and dss are not supported on OS X.
                darwin_unsupported = ('--enable-prof' in config_opts or \
                  'dss:primary' in malloc_conf_opts)
                if (uname == 'Linux' and linux_supported) \
                  or (not linux_supported and (uname != 'Darwin' or \
                  not darwin_unsupported)):
                    print """cat <<EOF > run_test_%(ind)d.sh
#!/bin/sh

set -e

abort() {
    echo "==> Error" >> run_test.log
    echo "Error; see run_tests.out/run_test_%(ind)d.out/run_test.log"
    exit 255 # Special exit code tells xargs to terminate.
}

# Environment variables are not supported.
run_cmd() {
    echo "==> \$@" >> run_test.log
    \$@ >> run_test.log 2>&1 || abort
}

echo "=> run_test_%(ind)d: %(config_line)s"
mkdir run_test_%(ind)d.out
cd run_test_%(ind)d.out

echo "==> %(config_line)s" >> run_test.log
%(config_line)s >> run_test.log 2>&1 || abort

run_cmd %(make_cmd)s all tests
run_cmd %(make_cmd)s check
run_cmd %(make_cmd)s distclean
EOF
chmod 755 run_test_%(ind)d.sh""" % {'ind': ind, 'config_line': config_line, 'make_cmd': make_cmd}
                    ind += 1

print 'for i in `seq 0 %(last_ind)d` ; do echo run_test_${i}.sh ; done | xargs -P %(nparallel)d -n 1 sh' % {'last_ind': ind-1, 'nparallel': nparallel}



# SPDX-License-Identifier: BSD-2-Clause
# Copyright (C) 2021 Intel Corporation.

import pytest
from python_framework import CMD_helper


class Test_max_bg_threads_env_var():

    cmd_helper = CMD_helper()
    fail_msg = "Test failed with:\n {0}"
    MAIN_THREAD = 1
    threads_limit = 0

    @property
    def min_threads(self):
        return self.MAIN_THREAD + self.threads_limit

    def run_test_binary(self):
        cmd_path = self.cmd_helper.get_command_path(
            '../environ_max_bg_threads_test')
        command = "MEMKIND_BACKGROUND_THREAD_LIMIT=" \
            f"{self.threads_limit} {cmd_path}"
        output, retcode = self.cmd_helper.execute_cmd(command)
        assert retcode != 1, \
            self.fail_msg.format(
                f"\nError: Execution of \'{command}\'"
                f" returns {retcode}. Output: {output}")
        return output, retcode

    def test_TC_MEMKIND_max_bg_threads_env_var_min(self):
        """This test checks if MEMKIND_BACKGROUND_THREAD_LIMIT
           environment variable limits the number of
           background threads spawned correctly."""
        self.threads_limit = 1
        output, retcode = self.run_test_binary()
        assert int(output) == self.min_threads, \
            self.fail_msg.format(
                f"Error: There should be {self.min_threads} threads running. "
                f"Counted threads: {output}.")

    @pytest.mark.parametrize("threads_limit", [0, 10])
    def test_TC_MEMKIND_max_bg_threads_env_var_multiple_threads(self,
                                                                threads_limit):
        """This test checks if MEMKIND_BACKGROUND_THREAD_LIMIT
           environment variable
           allows spawning of multiple threads."""
        output, threads_count = self.run_test_binary()
        assert int(output) > self.min_threads, \
            self.fail_msg.format(
                f"Error: There should be more than"
                f" {self.min_threads} threads running. "
                f"Counted threads: {output}.")

    def test_TC_MEMKIND_max_bg_threads_env_var_err(self):
        """This test checks if a negative value
           MEMKIND_BACKGROUND_THREAD_LIMIT environment variable is handled
           correctly."""
        self.threads_limit = -1
        output, retcode, threads_count = self.run_test_binary()
        assert retcode == 134, \
            self.fail_msg.format(
                "Error: Negative value of MEMKIND_BACKGROUND_THREAD_LIMIT"
                " should not be handled. "
                f"Counted threads: {threads_count}. Output: {output}")



# SPDX-License-Identifier: BSD-2-Clause
# Copyright (C) 2017 - 2021 Intel Corporation.

from python_framework.cmd_helper import CMD_helper
import os
import itertools

"""
Huge_page_organizer sets hugepages per NUMA node and
restores initial setup of hugepages. It writes and
reads from the same file, so using Huge_page_organizer with parallel
execution may cause undefined behaviour.
"""


class Huge_page_organizer(object):

    cmd_helper = CMD_helper()
    path = "/sys/devices/system/node/node{0}/hugepages/" \
           "hugepages-2048kB/nr_hugepages"
    restore_values = []

    def __restore(self):
        """Restore initial setup of hugepages."""
        for node_id, restore_value in enumerate(self.restore_values):
            self.__set_nr_hugepages(node_id, restore_value)

    def __get_nr_hugepages(self, node_id):
        """Return number of hugepages on given node
           or None if given node is not configured."""
        if not os.path.isfile(self.path.format(node_id)):
            return None
        with open(self.path.format(node_id), "r") as f:
            return int(f.read())

    def __set_nr_hugepages(self, node_id, nr_hugepages):
        """Set hugepages on given node to given number.
           Return True on success, False otherwise."""
        command = 'sh -c "echo {0} >> {1}"'.format(
            nr_hugepages, self.path.format(node_id))
        output, retcode = self.cmd_helper.execute_cmd(command, sudo=True)
        if retcode:
            return False
        return self.__get_nr_hugepages(node_id) is nr_hugepages

    def __init__(self, nr_hugepages_per_node):
        """Save current hugepages setup and set hugepages per NUMA node."""
        for node_id in itertools.count():
            nr_hugepages = self.__get_nr_hugepages(node_id)
            if nr_hugepages is None:
                break
            self.restore_values.append(nr_hugepages)
            retcode = self.__set_nr_hugepages(node_id, nr_hugepages_per_node)
            if not retcode:
                self.__restore()
            assert retcode, "Error: Could not set the requested"\
                            " amount of hugepages."

    def __del__(self):
        """Call __restore() function."""
        self.__restore()



# SPDX-License-Identifier: BSD-2-Clause
# Copyright (C) 2016 - 2021 Intel Corporation.

import matplotlib.pyplot as plt
import numpy as np
import numpy.ma as ma
import os
from shutil import rmtree

files = ('alloctest_hbw.txt', 'alloctest_glibc.txt',
         'alloctest_tbb.txt', 'alloctest_pmem.txt')
legend = ('avg hbw', 'avg glibc', 'avg tbb', 'avg pmem', 'first operation')
colors = ('red', 'green', 'blue', 'black')
first_operation_color = 'yellow'

threads_values = ('1', '2', '4', '8', '16', '18', '36')

small_sizes_values = ('1', '4', '16')
medium_sizes_values = ('64', '256')
big_sizes_values = ('1024', '4096', '16384')
sizes_values = (small_sizes_values, medium_sizes_values, big_sizes_values)
sizes_display = ('small', 'medium', 'big')

operations = ('Allocation', 'Free', 'Total')

iterations = 1000
output_directory = './plots/'

threads_index = np.arange(len(threads_values))
small_sizes_index = np.arange(len(small_sizes_values))
medium_sizes_index = np.arange(len(medium_sizes_values))
big_sizes_index = np.arange(len(big_sizes_values))
sizes_index = (small_sizes_index, medium_sizes_index, big_sizes_index)

# 3D and 2D plots needs different width,
# so that bars do not cover each other in 3D
bar_width_2D = 0.2
bar_width_3D = 0.1

# return times (total, allocation,
# free, first allocation, first free) as columns 2 - 6


def return_times(entry):
    return entry[:, 2], entry[:, 3], entry[:, 4], entry[:, 5], entry[:, 6]

# initialize axis with labels, ticks and values


def init_axis(fig, suptitle, subplot_projection,
              x_label, x_ticks, x_values, y_label, y_ticks, y_values, z_label):
    assert fig is not None
    fig.clf()
    fig.suptitle(suptitle)
    if subplot_projection:
        ax = fig.add_subplot(111, projection=subplot_projection)
    else:
        ax = fig.add_subplot(111)
    assert ax is not None
    if x_label:
        ax.set_xlabel(x_label)
        if x_ticks is not None and x_values is not None:
            ax.set_xticks(x_ticks)
            ax.set_xticklabels(x_values)
    if y_label:
        ax.set_ylabel(y_label)
        if y_ticks is not None and y_values is not None:
            ax.set_yticks(y_ticks)
            ax.set_yticklabels(y_values)
    if z_label:
        ax.set_zlabel(z_label)
    return ax


def draw_bar(ax, show_first_operation, x, y, time, init_time, draw_color):
    assert ax is not None
    if y is None:
        if show_first_operation is True:
            mask_time = ma.where(time >= init_time)
            mask_init_time = ma.where(init_time >= time)
            ax.bar(x[mask_time], time[mask_time],
                   width=bar_width, color=draw_color)
            ax.bar(x, init_time,
                   width=bar_width, color=first_operation_color)
            ax.bar(x[mask_init_time],
                   time[mask_init_time], width=bar_width, color=draw_color)
        else:
            ax.bar(x, time, width=bar_width, color=draw_color)
    else:
        if show_first_operation is True:
            mask_time = ma.where(time >= init_time)
            mask_init_time = ma.where(init_time >= time)
            ax.bar(x[mask_time], (time - init_time)[mask_time], y[mask_time],
                   bottom=init_time[mask_time], zdir='y',
                   width=bar_width, color=draw_color)
            ax.bar(x[mask_init_time], (init_time - time)[mask_init_time],
                   y[mask_init_time], bottom=time[mask_init_time],
                   zdir='y', width=bar_width, color=first_operation_color)
            ax.bar(x[mask_init_time], time[mask_init_time], y[mask_init_time],
                   zdir='y', width=bar_width, color=draw_color)
            ax.bar(x[mask_time], init_time[mask_time], y[mask_time],
                   zdir='y', width=bar_width, color=first_operation_color)
        else:
            ax.bar(x, time, y, zdir='y', width=bar_width, color=draw_color)


def save_plot(show_first_operation, subfolder, filename):
    if show_first_operation:
        path = os.path.join(output_directory, "with_first_operation")
    else:
        path = os.path.join(output_directory, "without_first_operation")
    if subfolder:
        path = os.path.join(path, subfolder)
    if not os.path.exists(path):
        os.mkdir(path)
    plt.savefig(os.path.join(path, filename))
    print("Saved file %s" % filename)


def load_data_from_files():
    data = []
    # load all files into data
    for f in files:
        assert os.path.isfile(f) is True
        data.append(np.loadtxt(f, comments='#'))
    return data


def set_bar_width_and_offsets(requested_width):
    bar_width = requested_width
    offsets = (0, bar_width, 2 * bar_width, 3 * bar_width)
    return bar_width, offsets


if os.path.exists(output_directory):
    rmtree(output_directory)
os.mkdir(output_directory)

data = load_data_from_files()

fig = plt.figure()

########################################
# Draw 3D plots (time, sizes, threads) #
########################################

bar_width, offsets = set_bar_width_and_offsets(bar_width_3D)

# for each size range (small, medium, big)
for size_values, size_index, size_display in zip(sizes_values,
                                                 sizes_index,
                                                 sizes_display):
    # show plots with and without first operation
    for show_first_operation in (True, False):
        # for each operation (allocation, free, total)
        for operation in operations:
            # add bar_width to each element of size_index
            ax = init_axis(fig, "%s time of %s sizes (%s iterations)" % (
                operation, size_display, iterations), '3d', 'size [kB]',
                size_index + (bar_width,) * len(size_index), size_values,
                'threads', threads_index, threads_values, 'time [ms]')
            legend_data = []
            # for each allocator (hbw, glibc, tbb, pmem)
            for entry, offset, draw_color in zip(data, offsets, colors):
                # remove all rows where column 1 (size) is not
                # in size_values (current size range)
                entry = entry[np.in1d(entry[:, 1],
                              np.array(size_values).astype(np.int))]
                # convert column 0 (threads values)
                # to thread index
                threads_col = np.array([threads_values.index(
                    str(n)) for n in map(int, entry[:, 0])])
                # convert column 1 (sizes values) to size index
                size_col = [size_values.index(
                    str(n)) for n in map(int, entry[:, 1])]
                # add offset to size index so that bars display near each other
                size_col = np.array(size_col) + offset
                (total_time_col, alloc_time_col, free_time_col,
                    first_alloc_time_col,
                    first_free_time_col) = return_times(entry)
                if operation == 'Allocation':
                    draw_bar(ax, show_first_operation, size_col,
                             threads_col, alloc_time_col,
                             first_alloc_time_col, draw_color)
                elif operation == 'Free':
                    draw_bar(ax, show_first_operation, size_col,
                             threads_col, free_time_col,
                             first_free_time_col, draw_color)
                elif operation == 'Total':
                    draw_bar(ax, show_first_operation, size_col,
                             threads_col, total_time_col,
                             first_alloc_time_col + first_free_time_col,
                             draw_color)
                legend_data.append(plt.Rectangle((0, 0), 1, 1, fc=draw_color))
            if show_first_operation is True:
                legend_data.append(plt.Rectangle(
                    (0, 0), 1, 1, fc=first_operation_color))
            ax.legend(legend_data, legend, loc='best')
            plt.grid()
            save_plot(show_first_operation, None,
                      "%s_time_of_%s_sizes_%s_iterations.png"
                      % (operation, size_display, iterations))

################################################
# Draw 2D plots (time, sizes, constant thread) #
################################################

bar_width, offsets = set_bar_width_and_offsets(bar_width_2D)

# for each size range (small, medium, big)
for size_values, size_index, size_display in zip(sizes_values,
                                                 sizes_index,
                                                 sizes_display):
    # show plots with and without first operation
    for show_first_operation in (True, False):
        for thread in threads_values:
            # for each operation (allocation, free, total)
            for operation in operations:
                # add bar_width to each element of size_index
                ax = init_axis(fig, "%s time of %s sizes with %s "
                               / "threads (%s operations)"
                               % (operation, size_display, thread, iterations),
                               None, 'size [kB]', size_index + (bar_width,)
                               * len(size_index), size_values,
                               'time [ms]', None, None, None)
                legend_data = []
                # for each allocator (hbw, glibc, tbb, pmem)
                for entry, offset, draw_color in zip(data, offsets, colors):
                    # remove all rows where column 1 (size)
                    # is not in size_values (current size range)
                    entry = entry[np.in1d(
                        entry[:, 1], np.array(size_values).astype(np.int))]
                    # remove all rows where column 0 (threads)
                    # is not equal to currently analyzed thread value
                    entry = entry[entry[:, 0] == int(thread)]
                    # convert column 0 (threads values) to thread index
                    threads_col = np.array([threads_values.index(
                        str(n)) for n in map(int, entry[:, 0])])
                    # convert column 1 (size values) to size index
                    size_col = [size_values.index(
                        str(n)) for n in map(int, entry[:, 1])]
                    # add offset to size index
                    # so that bars display near each other
                    size_col = np.array(size_col) + offset
                    (total_time_col, alloc_time_col, free_time_col,
                        first_alloc_time_col,
                        first_free_time_col) = return_times(entry)
                    if operation == 'Allocation':
                        draw_bar(ax, show_first_operation, size_col, None,
                                 alloc_time_col, first_alloc_time_col,
                                 draw_color)
                    elif operation == 'Free':
                        draw_bar(ax, show_first_operation, size_col, None,
                                 free_time_col, first_free_time_col,
                                 draw_color)
                    elif operation == 'Total':
                        draw_bar(ax, show_first_operation, size_col, None,
                                 total_time_col,
                                 first_alloc_time_col + first_free_time_col,
                                 draw_color)
                    legend_data.append(plt.Rectangle(
                        (0, 0), 1, 1, fc=draw_color))
                if show_first_operation is True:
                    legend_data.append(plt.Rectangle(
                        (0, 0), 1, 1, fc=first_operation_color))
                ax.legend(legend_data, legend, loc='best')
                plt.grid()
                save_plot(show_first_operation, "constant_thread",
                          "%s_time_of_%s_sizes_with_%s_threads_%s_iterations"
                          / ".png" % (operation,
                                      size_display, thread, iterations))

################################################
# Draw 2D plots (time, threads, constant size) #
################################################

# for each size range (small, medium, big)
for size_values, size_index, size_display in zip(sizes_values,
                                                 sizes_index,
                                                 sizes_display):
    # show plots with and without first operation
    for show_first_operation in (True, False):
        for size in size_values:
            # for each operation (allocation, free, total)
            for operation in operations:
                # add bar_width to each element of threads_index
                ax = init_axis(fig, "%s time of %s kB (%s operations)"
                               % (operation, size, iterations), None,
                               'threads', threads_index + (bar_width,)
                               * len(threads_index), threads_values,
                               'time [ms]', None, None,
                               None)
                legend_data = []
                # for each allocator (hbw, glibc, tbb, pmem)
                for entry, offset, draw_color in zip(data, offsets, colors):
                    # remove all rows where column 1 (size)
                    # is not in size_values (current size range)
                    entry = entry[np.in1d(entry[:, 1],
                                          np.array(
                                              size_values).astype(np.int))]
                    # remove all rows where column 1 (size)
                    # is not equal to currently analyzed size value
                    entry = entry[entry[:, 1] == int(size)]
                    # convert column 0 (threads values) to thread index
                    threads_col = np.array(
                        [threads_values.index(
                            str(n)) for n in map(int, entry[:, 0])])
                    # add offset to thread index
                    # so that bars display near each other
                    threads_col = np.array(threads_col) + offset
                    # convert column 1 (size values) to size index
                    size_col = [size_values.index(
                        str(n)) for n in map(int, entry[:, 1])]
                    (total_time_col, alloc_time_col, free_time_col,
                        first_alloc_time_col,
                        first_free_time_col) = return_times(entry)
                    if operation == 'Allocation':
                        draw_bar(ax, show_first_operation, threads_col, None,
                                 alloc_time_col, first_alloc_time_col,
                                 draw_color)
                    elif operation == 'Free':
                        draw_bar(ax, show_first_operation, threads_col, None,
                                 free_time_col, first_free_time_col,
                                 draw_color)
                    elif operation == 'Total':
                        draw_bar(ax, show_first_operation, threads_col, None,
                                 total_time_col,
                                 first_alloc_time_col + first_free_time_col,
                                 draw_color)
                    legend_data.append(plt.Rectangle(
                        (0, 0), 1, 1, fc=draw_color))
                if show_first_operation is True:
                    legend_data.append(plt.Rectangle(
                        (0, 0), 1, 1, fc=first_operation_color))
                ax.legend(legend_data, legend, loc='best')
                plt.grid()
                save_plot(show_first_operation, "constant_size",
                          "%s_time_of_%s_kB_%s_operations.png"
                          % (operation, size, iterations))



# SPDX-License-Identifier: BSD-2-Clause
# Copyright (C) 2016 - 2021 Intel Corporation.

import os
import tempfile
import subprocess


class CMD_helper(object):

    def execute_cmd(self, command, sudo=False):
        if sudo:
            command = "sudo {0}".format(command)
        # Initialize temp file for stdout. Will be removed when closed.
        outfile = tempfile.SpooledTemporaryFile()
        try:
            # Invoke process
            p = subprocess.Popen(
                command, stdout=outfile, stderr=subprocess.STDOUT, shell=True)
            p.communicate()
            # Read stdout from file
            outfile.seek(0)
            stdout = outfile.read()
        except subprocess.CalledProcessError:
            raise
        finally:
            # Make sure the file is closed
            outfile.close()
        retcode = p.returncode
        return stdout.decode("utf-8"), retcode

    def get_command_path(self, binary):
        """Get the path to the binary."""
        path = os.path.dirname(os.path.abspath(__file__))
        return os.path.join(path, binary)



# SPDX-License-Identifier: BSD-2-Clause
# Copyright (C) 2021 Intel Corporation.

from python_framework.cmd_helper import CMD_helper
import pytest


class Test_hbw_env_var(object):
    environ_err_threshold_test = "../environ_err_hbw_threshold_test"
    cmd_helper = CMD_helper()

    def run_test(self, threshold, kind):
        env = "MEMKIND_HBW_THRESHOLD=" + str(threshold)
        bin_path = self.cmd_helper.get_command_path(
            self.environ_err_threshold_test)
        command = " ".join([env, bin_path, kind])
        output, retcode = self.cmd_helper.execute_cmd(command)
        print(output)
        fail_msg = (f"Test failed with error: \nExecution of: \'{command}\'"
                    f" returns: {retcode} \noutput: {output}")
        assert retcode == 0, fail_msg

    @pytest.mark.parametrize("kind", ["MEMKIND_HBW", "MEMKIND_HBW_ALL"])
    def test_TC_MEMKIND_hbw_threshold_default_value(self, kind):
        threshold = 204800
        self.run_test(threshold, kind)

    @pytest.mark.parametrize("kind", ["MEMKIND_HBW", "MEMKIND_HBW_ALL"])
    def test_TC_MEMKIND_hbw_threshold_negative_value(self, kind):
        threshold = -1
        self.run_test(threshold, kind)

    @pytest.mark.parametrize("kind", ["MEMKIND_HBW", "MEMKIND_HBW_ALL"])
    def test_TC_MEMKIND_hbw_threshold_low_value(self, kind):
        threshold = 1
        self.run_test(threshold, kind)

    @pytest.mark.parametrize("kind", ["MEMKIND_HBW", "MEMKIND_HBW_ALL"])
    def test_TC_MEMKIND_hbw_threshold_high_value(self, kind):
        threshold = 1024 * 1024 * 1024
        self.run_test(threshold, kind)



# SPDX-License-Identifier: BSD-2-Clause
# Copyright (C) 2021 Intel Corporation.

import os
import pytest
import re
import subprocess

from python_framework import cmd_helper

MEMKIND_PMEM_MIN_SIZE = 1024 * 1024 * 16


class Helper(object):
    # NOTE: this script should be called from the root of memkind repository
    ld_preload_env = "LD_PRELOAD=tiering/.libs/libmemtier.so"
    # TODO create separate, parametrized binary that could be used for testing
    # instead of using ls here
    bin_path = "ls"
    cmd = cmd_helper.CMD_helper()

    log_prefix = "MEMKIND_MEM_TIERING_"
    log_debug_prefix = log_prefix + "LOG_DEBUG: "
    log_error_prefix = log_prefix + "LOG_ERROR: "
    log_info_prefix = log_prefix + "LOG_INFO: "

    kind_name_dict = {
        'DRAM': 'memkind_default',
        'FS_DAX': 'FS-DAX',
        'DAX_KMEM': 'memkind_dax_kmem'}

    mem_tiers_env_var = "MEMKIND_MEM_TIERS"
    mem_thresholds_env_var = "MEMKIND_MEM_THRESHOLDS"

    # POLICY_STATIC_RATIO is a policy used in tests that have to set a
    # valid policy but don't test anything related to allocation policies
    default_policy = "POLICY:STATIC_RATIO"

    @staticmethod
    def bytes_from_str(bytes_str):
        multiplier = 1
        if bytes_str[-1] == 'K':
            multiplier = 1 << 10
        elif bytes_str[-1] == 'M':
            multiplier = 1 << 20
        elif bytes_str[-1] == 'G':
            multiplier = 1 << 30

        return str(multiplier * int(bytes_str[:-1]))

    def check_fs_dax_support(self):
        fs_dax_path = os.environ.get('PMEM_PATH', '/tmp').rstrip("/")
        with open('/proc/mounts', 'r') as f:
            for line in f.readlines():
                if 'dax' in line and fs_dax_path in line:
                    return True
        return False

    def check_kmem_dax_support(self):
        try:
            subprocess.check_call("./memkind-auto-dax-kmem-nodes")
        except subprocess.CalledProcessError:
            return False
        else:
            return True

    def get_ld_preload_cmd_output(self, tiers_config, thresholds_config=None,
                                  log_level=None, negative_test=False):
        log_level_env = self.log_prefix + "LOG_LEVEL=" + log_level \
            if log_level else ""
        tiers_env = self.mem_tiers_env_var + '="' + tiers_config + '"'
        thresholds_env = self.mem_thresholds_env_var + '="' + \
            thresholds_config + '"' if thresholds_config else ""
        command = " ".join([self.ld_preload_env, log_level_env,
                            tiers_env, thresholds_env, self.bin_path])
        output, retcode = self.cmd.execute_cmd(command)
        fail_msg = "Execution of: '" + command + \
            "' returns: " + str(retcode) + "\noutput: " + output

        if negative_test:
            assert retcode != 0, fail_msg
        else:
            assert retcode == 0, fail_msg
        return output.splitlines()

    def get_default_cmd_output(self):
        output, retcode = self.cmd.execute_cmd(self.bin_path)

        assert retcode == 0, \
            "Execution of: '" + self.bin_path + \
            "' returns: " + str(retcode) + "\noutput: " + output
        return output.splitlines()


class Test_tiering_log(Helper):

    def test_utils_init(self):
        default_output = self.get_default_cmd_output()

        output = self.get_ld_preload_cmd_output(
            "KIND:DRAM,RATIO:1;" + self.default_policy,
            log_level="1")

        assert output[0] == self.log_info_prefix + \
            "Memkind memtier lib loaded!", "Bad init message"

        # check if rest of output from command is unchanged
        rest_output = output[1:]
        assert rest_output == default_output, "Bad command output"

    def test_utils_log_level_default(self):
        default_output = self.get_default_cmd_output()

        output = self.get_ld_preload_cmd_output(
            "KIND:DRAM,RATIO:1;" + self.default_policy, log_level="0")

        assert output == default_output, "Bad command output"

    def test_utils_log_level_not_set(self):
        default_output = self.get_default_cmd_output()

        output = self.get_ld_preload_cmd_output(
            "KIND:DRAM,RATIO:1;" + self.default_policy)

        assert output == default_output, "Bad command output"

    def test_utils_log_level_debug(self):
        re_hex_or_nil = r"((0[xX][a-fA-F0-9]+)|\(nil\))"
        re_log_debug_valid_messages = [
            self.log_debug_prefix
            + r"malloc\(\d+\) = " + re_hex_or_nil + r"$",
            self.log_debug_prefix
            + r"realloc\(" + re_hex_or_nil + r", \d+\) = "
            + re_hex_or_nil + r"$",
            self.log_debug_prefix
            + r"calloc\(\d+, \d+\) = " + re_hex_or_nil + r"$",
            self.log_debug_prefix
            + r"memalign\(\d+, \d +\) = " + re_hex_or_nil + r"$",
            self.log_debug_prefix + r"free\(" + re_hex_or_nil + r"\)$",
            self.log_debug_prefix + r"kind_name: \w+$",
            self.log_debug_prefix + r"pmem_path: .*$",  # TODO add path re
            self.log_debug_prefix + r"pmem_size: \d+$",
            self.log_debug_prefix + r"ratio_value: \d+$",
            self.log_debug_prefix + r"policy: \w+$",
        ]

        default_output = self.get_default_cmd_output()

        output = self.get_ld_preload_cmd_output(
            "KIND:DRAM,RATIO:1;" + self.default_policy,
            log_level="2")

        assert output[0] == self.log_debug_prefix + \
            "Setting log level to: 2", "Bad init message"

        assert output[1] == self.log_prefix + "LOG_INFO: " + \
            "Memkind memtier lib loaded!", "Bad init message"

        # extract from the output all lines starting with
        # "MEMKIND_MEM_TIERING_LOG" prefix and check if they are correct
        log_output = [line for line in output[2:]
                      if line.startswith(self.log_prefix)]
        for log_line in log_output:
            log_line_valid = any(re.match(pattern, log_line) for pattern in
                                 re_log_debug_valid_messages)
            assert log_line_valid, "Bad log message format: " + log_line

        # finally check if rest of output from command is unchanged
        output = [line for line in output if line not in log_output][2:]
        assert output == default_output, "Bad ls output"

    def test_utils_log_level_negative(self):
        log_level_env = self.log_prefix + "LOG_LEVEL=-1"
        command = " ".join([self.ld_preload_env, log_level_env,
                            self.bin_path])
        output_level_neg, _ = self.cmd.execute_cmd(command)

        assert output_level_neg.splitlines()[0] == \
            self.log_error_prefix + "Wrong value of " + \
            self.log_prefix + "LOG_LEVEL=-1", "Bad init message"

    def test_utils_log_level_too_high(self):
        log_level_env = self.log_prefix + "LOG_LEVEL=4"
        command = " ".join([self.ld_preload_env, log_level_env,
                            self.bin_path])
        output_level_neg, _ = self.cmd.execute_cmd(command)

        assert output_level_neg.splitlines()[0] == \
            self.log_error_prefix + "Wrong value of " + \
            self.log_prefix + "LOG_LEVEL=4", "Bad init message"


class Test_memkind_log(Helper):

    log_info_prefix = "MEMKIND_INFO: "

    re_hex_or_nil = r"((0[xX][a-fA-F0-9]+)|\(nil\))"
    re_log_valid_messages = [
        log_info_prefix
        + r"malloc\(\d+\) = " + re_hex_or_nil + r" from kind: ",
        log_info_prefix
        + r"realloc\(" + re_hex_or_nil + r", \d+\) = "
        + re_hex_or_nil + r" from kind: ",
        log_info_prefix
        + r"calloc\(\d+, \d+\) = " + re_hex_or_nil + r" from kind: ",
        log_info_prefix
        + r"memalign\(\d+, \d +\) = " + re_hex_or_nil + r" from kind: ",
        log_info_prefix + r"free\(" + re_hex_or_nil + r"\) from kind: ",
    ]

    def test_DRAM_only(self):
        memkind_debug_env = "MEMKIND_DEBUG=1"
        tiering_cfg_env = "KIND:DRAM,RATIO:1;" + \
            self.default_policy

        command = " ".join([self.ld_preload_env, tiering_cfg_env,
                            memkind_debug_env, self.bin_path])
        self.cmd.execute_cmd(command)

        # TODO enable this check after implementation of decorators in
        # memkind_memtier.c
        """
        kind_name = "memkind_default$"
        output = [l for l in output if l.startswith(self.log_info_prefix)]

        for log_line in output:
            log_line_valid = any(re.match(pattern + kind_name, log_line)
                                 for pattern in self.re_log_valid_messages)
            assert log_line_valid, "Bad log message format: " + log_line
        """


class Test_tiering_config_env(Helper):

    def test_DRAM_ratio_first(self):
        self.get_ld_preload_cmd_output(
            "RATIO:1,KIND:DRAM;" + self.default_policy,
            log_level="2")

    @pytest.mark.parametrize("param_name",
                             ["fafes", "fdas3rfs43", "%#3^"])
    def test_bad_param_name(self, param_name):
        self.get_ld_preload_cmd_output(
            "RATIO:1," + param_name + ":DRAM;" + self.default_policy,
            negative_test=True)

    @pytest.mark.parametrize("param_name",
                             ["PMEM_SIZE_LIMIT:1G", "PATH:/tmp/"])
    def test_unsupported_param(self, param_name):
        self.get_ld_preload_cmd_output(
            "KIND:DRAM," + param_name + ",RATIO:1;" + self.default_policy,
            negative_test=True)

    @pytest.mark.parametrize("fsdax_cfg",
                             ["KIND:FS_DAX,PATH:/tmp/,PMEM_SIZE_LIMIT:10G,"
                              "RATIO:1",
                              "PATH:/tmp/,KIND:FS_DAX,PMEM_SIZE_LIMIT:10G,"
                              "RATIO:1",
                              "KIND:FS_DAX,PMEM_SIZE_LIMIT:10G,PATH:/tmp/,"
                              "RATIO:1",
                              "RATIO:1,PMEM_SIZE_LIMIT:10G,KIND:FS_DAX,"
                              "PATH:/tmp/"])
    def test_FS_DAX_various_param_order(self, fsdax_cfg):
        self.get_ld_preload_cmd_output(
            fsdax_cfg + ";" + self.default_policy,
            log_level="2")

    def test_no_config(self):
        self.get_ld_preload_cmd_output('', negative_test=True)

    @pytest.mark.parametrize("order", [0, 1])
    def test_config_string_order(self, order):
        tiers_str = "KIND:DRAM,RATIO:1;" + \
            "KIND:FS_DAX,PATH:/tmp/,PMEM_SIZE_LIMIT:100M,RATIO:4;" + \
            "POLICY:DYNAMIC_THRESHOLD"
        threshold_str = "INIT_VAL:1K,MIN_VAL:64,MAX_VAL:2K"

        log_level_env = self.log_prefix + "LOG_LEVEL=2"
        tiers_env = self.mem_tiers_env_var + '="' + tiers_str + '"'
        thresholds_env = self.mem_thresholds_env_var + '="' + \
            threshold_str + '"'

        first_env = tiers_env if order else thresholds_env
        second_env = thresholds_env if order else tiers_env

        command = " ".join([self.ld_preload_env, log_level_env, first_env,
                            second_env, self.bin_path])

        output, retcode = self.cmd.execute_cmd(command)
        fail_msg = "Execution of: '" + command + \
            "' returns: " + str(retcode) + "\noutput: " + output
        assert retcode == 0, fail_msg

    @pytest.mark.parametrize("ratio", ["1", "1000", "4294967295"])
    def test_DRAM_ratio(self, ratio):
        self.get_ld_preload_cmd_output(
            "KIND:DRAM,RATIO:" + ratio + ";" + self.default_policy,
            log_level="2")

    @pytest.mark.parametrize("policy", ["POLICY:STATIC_RATIO"])
    def test_DRAM_policy(self, policy):
        ratio = "1"
        self.get_ld_preload_cmd_output(
            "KIND:DRAM,RATIO:" + ratio + ";" + policy,
            log_level="2")

    def test_KMEM_DAX(self):
        if not self.check_kmem_dax_support():
            pytest.skip("KMEM DAX is not configured")
        self.get_ld_preload_cmd_output(
            "KIND:DRAM,RATIO:1;KIND:KMEM_DAX,RATIO:100;" + self.default_policy)

    @pytest.mark.parametrize("policy", ["POLICY:STATIC_RATIO",
                                        "POLICY:DYNAMIC_THRESHOLD"])
    def test_KMEM_DAX_multiple(self, policy):
        self.get_ld_preload_cmd_output(
            "KIND:DRAM,RATIO:1;KIND:KMEM_DAX,RATIO:10;KIND:KMEM_DAX,RATIO:8;"
            + policy, negative_test=True)

    @pytest.mark.parametrize("pmem_size", ["0", str(MEMKIND_PMEM_MIN_SIZE),
                                           "18446744073709551615"])
    def test_FSDAX(self, pmem_size):
        self.get_ld_preload_cmd_output(
            "KIND:FS_DAX,PATH:/tmp/,PMEM_SIZE_LIMIT:" + pmem_size + ",RATIO:1;"
            + self.default_policy, log_level="2")

    @pytest.mark.parametrize("pmem_size",
                             ["1073741824", "1048576K", "1024M", "1G"])
    def test_FSDAX_pmem_size_with_suffix(self, pmem_size):
        self.get_ld_preload_cmd_output(
            "KIND:FS_DAX,PATH:/tmp/,PMEM_SIZE_LIMIT:" + pmem_size + ",RATIO:1;"
            + self.default_policy, log_level="2")

    @pytest.mark.parametrize("pmem_size",
                             ["-1", "-4294967295", "-18446744073709551615",
                              "18446744073709551616"])
    def test_FSDAX_pmem_size_outside_limits(self, pmem_size):
        self.get_ld_preload_cmd_output(
            "KIND:FS_DAX,PATH:/tmp/,PMEM_SIZE_LIMIT:" + pmem_size + ",RATIO:1;"
            + self.default_policy, negative_test=True)

    @pytest.mark.parametrize("pmem_size",
                             ["1", str(MEMKIND_PMEM_MIN_SIZE - 1)])
    def test_FSDAX_pmem_size_too_small(self, pmem_size):
        self.get_ld_preload_cmd_output(
            "KIND:FS_DAX,PATH:/tmp/,PMEM_SIZE_LIMIT:" + pmem_size + ",RATIO:1;"
            + self.default_policy, negative_test=True)

    @pytest.mark.parametrize("pmem_size",
                             ["18446744073709551615K", "18446744073709551615M",
                              "18446744073709551615G"])
    def test_FSDAX_pmem_size_with_suffix_too_big(self, pmem_size):
        self.get_ld_preload_cmd_output(
            "KIND:FS_DAX,PATH:/tmp/,PMEM_SIZE_LIMIT" + pmem_size + ",RATIO:1;"
            + self.default_policy, negative_test=True)

    def test_FSDAX_no_size(self):
        self.get_ld_preload_cmd_output(
            "KIND:FS_DAX", negative_test=True)

    @pytest.mark.parametrize("pmem_size", ["as", "10K1", "M", "M2", "10KM"])
    def test_FSDAX_wrong_size(self, pmem_size):
        self.get_ld_preload_cmd_output(
            "KIND:FS_DAX,PATH:/tmp/,PMEM_SIZE_LIMIT:" + pmem_size + ",RATIO:1;"
            + self.default_policy, negative_test=True)

    def test_FSDAX_check_only_fs_dax(self):
        pmem_path = os.environ.get('PMEM_PATH')
        if not self.check_fs_dax_support():
            pytest.skip("Missing FS DAX mounted on" + pmem_path)
        self.get_ld_preload_cmd_output(
            "KIND:FS_DAX,PATH:" + pmem_path + ",PMEM_SIZE_LIMIT:1G,RATIO:1;"
            + self.default_policy, log_level="2")

    def test_FSDAX_negative_ratio(self):
        self.get_ld_preload_cmd_output(
            "KIND:FS_DAX,PATH:/tmp/,PMEM_SIZE_LIMIT:10G,RATIO:-1;"
            + self.default_policy, negative_test=True)

    def test_FSDAX_wrong_ratio(self):
        self.get_ld_preload_cmd_output(
            "KIND:FS_DAX,PATH:/tmp/,PMEM_SIZE_LIMIT:10G,RATIO:a;"
            + self.default_policy, negative_test=True)

    def test_FSDAX_no_path(self):
        self.get_ld_preload_cmd_output(
            "KIND:FS_DAX,PMEM_SIZE_LIMIT:10G,RATIO:1;"
            + self.default_policy, negative_test=True)

    def test_FSDAX_no_max_size(self):
        self.get_ld_preload_cmd_output(
            "KIND:FS_DAX,PATH:/tmp/,RATIO:1;"
            + self.default_policy, log_level="2")

    def test_class_not_defined(self):
        self.get_ld_preload_cmd_output("2", negative_test=True)

    def test_bad_ratio(self):
        self.get_ld_preload_cmd_output(
            "KIND:DRAM,RATIO:A;" + self.default_policy, negative_test=True)

    def test_no_ratio(self):
        self.get_ld_preload_cmd_output(
            "KIND:DRAM;" + self.default_policy, negative_test=True)

    def test_bad_class(self):
        self.get_ld_preload_cmd_output(
            "KIND:a1b2,RATIO:10;" + self.default_policy, negative_test=True)

    def test_negative_ratio(self):
        self.get_ld_preload_cmd_output(
            "KIND:DRAM,RATIO:-1;" + self.default_policy, negative_test=True)

    def test_double_ratio(self):
        self.get_ld_preload_cmd_output(
            "RATIO:2,KIND:DRAM,RATIO:1;" + self.default_policy,
            negative_test=True)

    def test_multiple_kind(self):
        self.get_ld_preload_cmd_output(
            "KIND:DRAM,RATIO:1,KIND:FS_DAX;" + self.default_policy,
            negative_test=True)

    @pytest.mark.parametrize("policy_str",
                             ["ABC", "5252", "1AB2C3", "#@%srfs"])
    def test_negative_unsupported_policy(self, policy_str):
        self.get_ld_preload_cmd_output(
            "KIND:DRAM,RATIO:1;POLICY:" + policy_str,
            negative_test=True)

    def test_negative_multiple_policies(self):
        self.get_ld_preload_cmd_output(
            "KIND:DRAM,RATIO:1;" + self.default_policy + ";"
            + self.default_policy, negative_test=True)

    def test_negative_policy_not_last(self):
        self.get_ld_preload_cmd_output(
            self.default_policy + ";KIND:DRAM,RATIO:1",
            negative_test=True)

    @pytest.mark.parametrize("config_str", ["", ",", ",,,", ":", "::"])
    def test_negative_config_str_invalid(self, config_str):
        self.get_ld_preload_cmd_output(
            config_str, negative_test=True)

    @pytest.mark.parametrize("query_str", [":", ":::"])
    def test_negative_query_str_invalid(self, query_str):
        self.get_ld_preload_cmd_output(
            "KIND" + query_str + "DRAM;" + self.default_policy,
            negative_test=True)

    @pytest.mark.parametrize("tier_str",
                             ["KIND:DRAM,RATIO:1",
                              "KIND:FS_DAX,PATH:/tmp/,PMEM_SIZE_LIMIT:100M,"
                              "RATIO:1",
                              "KIND:DRAM,RATIO:1;KIND:FS_DAX,PATH:/tmp/,"
                              "PMEM_SIZE_LIMIT:100M,RATIO:1"])
    def test_negative_no_policy(self, tier_str):
        self.get_ld_preload_cmd_output(
            tier_str, negative_test=True)

    @pytest.mark.parametrize("policy",
                             ["POLICY:STATIC_RATIO",
                              "POLICY:DYNAMIC_THRESHOLD"])
    def test_multiple_tiers(self, policy):
        assert os.access("/mnt", os.W_OK | os.X_OK), \
            "Write and execute permissions to the /mnt directory "
        "are required for this test"
        self.get_ld_preload_cmd_output(
            "KIND:DRAM,RATIO:1;KIND:FS_DAX,PATH:/tmp/,PMEM_SIZE_LIMIT:100M,"
            "RATIO:4;KIND:FS_DAX,PATH:/mnt,PMEM_SIZE_LIMIT:150M,RATIO:8;"
            + policy, log_level="2")

    @pytest.mark.parametrize("thresholds",
                             ["INIT_VAL:1K,MIN_VAL:64,MAX_VAL:2K",
                              "INIT_VAL:100,MIN_VAL:1,MAX_VAL:1000",
                              "MIN_VAL:1,MAX_VAL:1000,INIT_VAL:100",
                              "MIN_VAL:1,INIT_VAL:100,MAX_VAL:1000",
                              "MAX_VAL:1000,MIN_VAL:1,INIT_VAL:100"])
    def test_two_tiers_thresholds(self, thresholds):
        self.get_ld_preload_cmd_output(
            "KIND:DRAM,RATIO:1;"
            + "KIND:FS_DAX,PATH:/tmp/,PMEM_SIZE_LIMIT:100M,RATIO:4;"
            + "POLICY:DYNAMIC_THRESHOLD", thresholds_config=thresholds,
            log_level="2")

    @pytest.mark.parametrize("thresholds",
                             ["::", "AAA", "A1B2C3", "AAA:BB", "AA:42",
                              "INIT_VAL:AAA", "INIT_VAL:1A2B", "INIT_VAL::",
                              "INIT_VAL:14,4.4", "INIT_VAL:111,INIT_VAL:222"])
    def test_negative_two_tiers_bad_thresholds(self, thresholds):
        self.get_ld_preload_cmd_output(
            "KIND:DRAM,RATIO:1;"
            + "KIND:FS_DAX,PATH:/tmp/,PMEM_SIZE_LIMIT:100M,RATIO:4;"
            + "POLICY:DYNAMIC_THRESHOLD", thresholds_config=thresholds,
            negative_test=True)

    @pytest.mark.parametrize("thresholds",
                             ["INIT_VAL:101,MIN_VAL:1,MAX_VAL:100",
                              "INIT_VAL:50,MIN_VAL:100,MAX_VAL:10",
                              "INIT_VAL:1,MIN_VAL:10,MAX_VAL:100"])
    def test_negative_two_tiers_bad_threshold_values(self, thresholds):
        self.get_ld_preload_cmd_output(
            "KIND:DRAM,RATIO:1;"
            + "KIND:FS_DAX,PATH:/tmp/,PMEM_SIZE_LIMIT:100M,RATIO:4;"
            + "POLICY:DYNAMIC_THRESHOLD", thresholds_config=thresholds,
            negative_test=True)

    def test_negative_thresholds_bad_policy(self):
        self.get_ld_preload_cmd_output(
            "KIND:DRAM,RATIO:1;"
            + "KIND:FS_DAX,PATH:/tmp/,PMEM_SIZE_LIMIT:100M,RATIO:4;"
            + "POLICY:STATIC_RATIO",
            thresholds_config="INIT_VAL:1K,MIN_VAL:64,MAX_VAL:2K",
            negative_test=True)

    @pytest.mark.parametrize("params",
                             ["INIT_VAL:1K,INIT_VAL:2K,MIN_VAL:64,MAX_VAL:3K",
                              "INIT_VAL:1K,MIN_VAL:64,MAX_VAL:2K,MAX_VAL:10K",
                              "INIT_VAL:1K,MIN_VAL:3K,MIN_VAL:64,MAX_VAL:2K",
                              "INIT_VAL:1K,MIN_VAL:64,MAX_VAL:2K,INIT_VAL:10K",
                              "INIT_VAL:1K,MIN_VAL:64,MAX_VAL:2K;"
                              "INIT_VAL:10K,MIN_VAL:3K,MAX_VAL:20K"])
    def test_negative_too_many_threshold_params(self, params):
        self.get_ld_preload_cmd_output(
            "KIND:DRAM,RATIO:1;"
            + "KIND:FS_DAX,PATH:/tmp/,PMEM_SIZE_LIMIT:100M,RATIO:4;"
            + "POLICY:STATIC_RATIO",
            thresholds_config=params,
            negative_test=True)

    @pytest.mark.parametrize("wrong_tier",
                             ["KIND:non_existent,PATH:/mnt,"
                              "PMEM_SIZE_LIMIT:150M,RATIO:8",
                              "KIND:FS_DAX,PATH:/non_existent,"
                              "PMEM_SIZE_LIMIT:150M,RATIO:8",
                              "KIND:FS_DAX,PATH:/mnt,PMEM_SIZE_LIMIT:1M,"
                              "RATIO:8",
                              "KIND:FS_DAX,PATH:/mnt,PMEM_SIZE_LIMIT:150M,"
                              "RATIO:0",
                              "KIND:DRAM,RATIO:1"])
    def test_multiple_tiers_wrong_tier(self, wrong_tier):
        assert os.access("/mnt", os.W_OK | os.X_OK), \
            "Write and execute permissions to the /mnt directory "
        "are required for this test"
        self.get_ld_preload_cmd_output(
            "KIND:DRAM,RATIO:1;KIND:FS_DAX,PATH:/tmp/,PMEM_SIZE_LIMIT:100M,"
            + "RATIO:4;" + wrong_tier + ";" + self.default_policy,
            log_level="2",
            negative_test=True)



# SPDX-License-Identifier: BSD-2-Clause
# Copyright (C) 2019 - 2021 Intel Corporation.

from distutils.spawn import find_executable
from python_framework.cmd_helper import CMD_helper
import os


class Test_dax_kmem_env_var(object):
    os.environ["PATH"] += (os.pathsep
                           + os.path.dirname(os.path.dirname(__file__)))
    binary_path = find_executable("memkind-auto-dax-kmem-nodes")
    environ_err_test = "../environ_err_dax_kmem_malloc_test"
    environ_err_positive_test = "../environ_err_dax_kmem_malloc_positive_test"
    expected_libnuma_warning = (
        "libnuma: Warning: node argument -1 is out of range\n\n")
    fail_msg = "Test failed with:\n {0}"
    cmd_helper = CMD_helper()

    def get_dax_kmem_nodes(self, nodemask=None):
        """ This function executes memkind function 'get_mbind_nodemask'
            and returns its output - comma-separated DAX_KMEM nodes """
        command = self.binary_path
        if (nodemask):
            command = "MEMKIND_DAX_KMEM_NODES={}".format(nodemask) + command
        output, retcode = self.cmd_helper.execute_cmd(command)
        assert retcode == 0, self.fail_msg.format(
            "\nError: Execution of \'{0}\' returns {1}, \noutput: {2}"
            .format(command, retcode, output))
        print("\nExecution of {} returns output {}".format(command, output))
        return output

    def test_TC_MEMKIND_dax_kmem_env_var_cmp_with_nodemask_default(self):
        """ This test checks whether dax_kmem_nodemask_default and
            dax_kmem_nodemask_env_variable have the same value """
        dax_kmem_nodemask_default = self.get_dax_kmem_nodes()
        dax_kmem_nodemask_env_variable = self.get_dax_kmem_nodes(
            dax_kmem_nodemask_default)
        assert dax_kmem_nodemask_default == dax_kmem_nodemask_env_variable, \
            self.fail_msg.format(
                "Error: Nodemask dax_kmem_nodemask_default ({0}) "
                "is not the same as nodemask "
                "dax_kmem_nodemask_env_variable ({1})".format(
                    dax_kmem_nodemask_default,
                    dax_kmem_nodemask_env_variable))

    def test_TC_MEMKIND_dax_kmem_env_var_negative_memkind_malloc(self):
        """ This test sets unsupported value of MEMKIND_DAX_KMEM_NODES,
            then tries to perform a successful allocation
            from DRAM using memkind_malloc() """
        command = ("MEMKIND_DAX_KMEM_NODES=-1 "
                   + self.cmd_helper.get_command_path(self.environ_err_test))
        output, retcode = self.cmd_helper.execute_cmd(command)
        assert retcode != 0, self.fail_msg.format(
            "\nError: Execution of: \'{0}\' returns: {1} \noutput: {2}"
            .format(command, retcode, output))
        assert self.expected_libnuma_warning == output, self.fail_msg.format(
            "Error: expected libnuma warning ({0}) "
            "was not found (output: {1})").format(
                self.expected_libnuma_warning, output)

    def test_TC_MEMKIND_dax_kmem_env_var_proper_memkind_malloc(self):
        """ This test checks if allocation is performed
            on persistent memory correctly """
        dax_kmem_nodemask_default = self.get_dax_kmem_nodes().strip()
        command = "MEMKIND_DAX_KMEM_NODES={} ".format(
            dax_kmem_nodemask_default) + self.cmd_helper.get_command_path(
                self.environ_err_positive_test)
        output, retcode = self.cmd_helper.execute_cmd(command)
        assert retcode == 0, self.fail_msg.format(
            "\nError: Execution of: \'{0}\' returns: {1} \noutput: {2}"
            .format(command, retcode, output))



# SPDX-License-Identifier: BSD-2-Clause
# Copyright (C) 2016 - 2021 Intel Corporation.

from python_framework.cmd_helper import CMD_helper


class Test_trace_mechanism(object):
    binary = "../trace_mechanism_test_helper"
    fail_msg = "Test failed with:\n {0}"
    debug_env = "MEMKIND_DEBUG=1 "
    cmd_helper = CMD_helper()

    def test_TC_MEMKIND_logging_MEMKIND_HBW(self):
        # This test executes trace_mechanism_test_helper and
        # test if MEMKIND_INFO message occurs while calling MEMKIND_HBW
        command = self.debug_env \
            + self.cmd_helper.get_command_path(self.binary) + " MEMKIND_HBW"
        print("Executing command: {0}".format(command))
        output, retcode = self.cmd_helper.execute_cmd(command, sudo=False)
        assert retcode == 0, self.fail_msg.format(
            "\nError: trace_mechanism_test_helper",
            " returned {0} \noutput: {1}".format(retcode, output))
        assert "MEMKIND_INFO: Initializing kind memkind_hbw." \
            in output, self.fail_msg.format(
                "\nError: trace mechanism in memkind doesn't show",
                " MEMKIND_INFO message \noutput: {0}").format(output)

    def test_TC_MEMKIND_2MBPages_logging_MEMKIND_HUGETLB(self):
        # This test executes trace_mechanism_test_helper and test
        # if MEMKIND_INFO message occurs while calling MEMKIND_HUGETLB
        command = self.debug_env \
            + self.cmd_helper.get_command_path(self.binary) \
            + " MEMKIND_HUGETLB"
        print("Executing command: {0}".format(command))
        output, retcode = self.cmd_helper.execute_cmd(command, sudo=False)
        assert retcode == 0, self.fail_msg.format(
            "\nError: trace_mechanism_test_helper",
            " returned {0} \noutput: {1}".format(retcode, output))
        assert "MEMKIND_INFO: Number of" in output, self.fail_msg.format(
            "\nError: trace mechanism in memkind doesn't show",
            " MEMKIND_INFO message \noutput: {0}").format(output)
        assert "MEMKIND_INFO: Overcommit limit for" \
            in output, self.fail_msg.format(
                "\nError: trace mechanism in memkind doesn't show",
                " MEMKIND_INFO message \n output: {0}").format(output)

    def test_TC_MEMKIND_logging_negative_MEMKIND_DEBUG_env(self):
        # This test executes trace_mechanism_test_helper and test
        # if setting MEMKIND_DEBUG to wrong value
        # causes MEMKIND_WARNING message
        command = "MEMKIND_DEBUG=-1 " \
            + self.cmd_helper.get_command_path(self.binary) + " MEMKIND_HBW"
        print("Executing command: {0}".format(command))
        output, retcode = self.cmd_helper.execute_cmd(command, sudo=False)
        assert retcode == 0, self.fail_msg.format(
            "\nError: trace_mechanism_test_helper",
            " returned {0} \noutput: {1}".format(retcode, output))
        assert "MEMKIND_WARNING: debug option" in output, self.fail_msg.format(
            "\nError: setting wrong MEMKIND_DEBUG environment variable",
            " doesn't show MEMKIND_WARNING \noutput: {0})").format(output)



# SPDX-License-Identifier: BSD-2-Clause
# Copyright (C) 2021 Intel Corporation.

import json
import pytest
from python_framework.cmd_helper import CMD_helper
import re


class Test_malloc_stats_print(object):

    cmd_helper = CMD_helper()
    bin_path = cmd_helper.get_command_path("../stats_print_test_helper")
    fail_msg = "Test failed with:\n {0}"
    bin_param = ""
    error_msg = "Error: {0} option not parsed correctly."

    def run_test_binary(self):
        command = " ".join([self.bin_path, self.bin_param])
        output, retcode = self.cmd_helper.execute_cmd(command)
        assert retcode == 0, \
            self.fail_msg.format(
                f"\nError: Execution of \'{command}\'",
                f" returns {retcode}. Output: {output}")
        return output, retcode

    @pytest.mark.parametrize("bin_param", ["default", "stdout", "no_write_cb"])
    def test_TC_MEMKIND_malloc_stats_print_check_output(self, bin_param):
        """ This test checks if there is output
            from malloc_stats_print() memkind API function """
        self.bin_param = bin_param
        output, _ = self.run_test_binary()
        assert output.endswith("--- End jemalloc statistics ---\n"), \
            "Error: Didn't get the expected output"\
            " from the '{self.bin_path}' binary."

    def test_TC_MEMKIND_malloc_stats_print_multi_opt_test(self):
        """ This test checks output from malloc_stats_print()
            memkind API function when options
            "MEMKIND_STAT_PRINT_JSON_FORMAT | MEMKIND_STAT_PRINT_OMIT_PER_ARENA
            | MEMKIND_STAT_PRINT_OMIT_EXTENT"
            are passed to the function """
        self.bin_param = "pass_opts"
        output, _ = self.run_test_binary()
        assert json.loads(output), \
            self.error_msg.format("MEMKIND_STAT_PRINT_JSON_FORMAT")
        assert "arenas[" not in output, \
               self.error_msg.format("MEMKIND_STAT_PRINT_OMIT_PER_ARENA")
        assert "extents:" not in output, \
               self.error_msg.format("MEMKIND_STAT_PRINT_OMIT_EXTENT")

    def test_TC_MEMKIND_malloc_stats_print_all_opts_test(self):
        """ This test checks output from malloc_stats_print()
            memkind API function when all possible options
            are passed to the function """
        self.bin_param = "all_opts"
        output, _ = self.run_test_binary()
        assert json.loads(output), \
            self.error_msg.format("MEMKIND_STAT_PRINT_JSON_FORMAT")
        assert "version" not in output, \
               self.error_msg.format("MEMKIND_STAT_PRINT_OMIT_GENERAL")
        assert "merged" not in output, \
               self.error_msg.format("MEMKIND_STAT_PRINT_OMIT_MERGED_ARENA")
        assert "destroyed" not in output, \
               self.error_msg.format(
                   "MEMKIND_STAT_PRINT_OMIT_DESTROYED_MERGED_ARENA")
        assert re.search(re.compile(r'(\d+)": {'), output) is None, \
            self.error_msg.format("MEMKIND_STAT_PRINT_OMIT_PER_ARENA")
        assert "\"bins\"" not in output, \
               self.error_msg.format(
                   "MEMKIND_STAT_PRINT_OMIT_PER_SIZE_CLASS_BINS")
        assert "lextents" not in output, \
               self.error_msg.format(
                   "MEMKIND_STAT_PRINT_OMIT_PER_SIZE_CLASS_LARGE")
        assert "mutex" not in output, \
               self.error_msg.format("MEMKIND_STAT_PRINT_OMIT_MUTEX")
        assert "\"extents\"" not in output, \
               self.error_msg.format("MEMKIND_STAT_PRINT_OMIT_EXTENT")

    def test_TC_MEMKIND_malloc_stats_print_negative_test(self):
        """ This test checks if there is no output from malloc_stats_print()
            memkind API function when wrong arguments
            are being passed to the function """
        self.bin_param = "negative_test"
        output, _ = self.run_test_binary()
        assert len(output) == 0, \
            "Error: There should be no output from"\
            " the '{self.bin_path}' binary."

    def test_TC_MEMKIND_malloc_stats_print_opts_negative_test(self):
        """ This test checks if there is a failure in parsing opts
            in malloc_stats_print() memkind API function
            when wrong options string is passed to the function """
        self.bin_param = "opts_negative_test"
        _, retcode = self.run_test_binary()
        assert retcode == 0, \
            f"Error: '{self.bin_path}' binary should return 0" \
            " indicating that parsing opts string failed."



# SPDX-License-Identifier: BSD-2-Clause
# Copyright (C) 2017 - 2021 Intel Corporation.

from distutils.spawn import find_executable
from python_framework.cmd_helper import CMD_helper
import os


class Test_hbw_detection(object):
    os.environ["PATH"] += os.pathsep + os.path.dirname(
        os.path.dirname(__file__))
    binary_path = find_executable("memkind-hbw-nodes")
    environ_err_test = "../environ_err_hbw_malloc_test"
    expected_libnuma_warning = "libnuma: Warning: node" \
        " argument -1 is out of range\n\n"
    fail_msg = "Test failed with:\n {0}"
    cmd_helper = CMD_helper()

    def get_hbw_nodes(self, nodemask=None):
        """ This function executes memkind function 'get_mbind_nodemask'
            and returns its output - comma-separated HBW nodes """
        command = self.binary_path
        if (nodemask):
            command = "MEMKIND_HBW_NODES={}".format(nodemask) + command
        output, retcode = self.cmd_helper.execute_cmd(command, sudo=False)
        assert retcode == 0, self.fail_msg.format(
            "\nError: Execution of: \'{0}\' returns: {1} \noutput: {2}".format(
                command, retcode, output))
        print("\nExecution of {} returns output {}".format(command, output))
        return output

    def test_TC_MEMKIND_hbw_detection_cmp_nodemask_default_and_env_var(self):
        """ This test checks whether hbw_nodemask_default
            and hbw_nodemask_env_variable has the same value """
        hbw_nodemask_default = self.get_hbw_nodes()
        hbw_nodemask_env_variable = self.get_hbw_nodes(hbw_nodemask_default)
        assert hbw_nodemask_default == hbw_nodemask_env_variable, \
            self.fail_msg.format(
                "Error: Nodemask hbw_nodemask_default ({0}) "
                "is not the same as nodemask "
                "hbw_nodemask_env_variable ({1})".format(
                    hbw_nodemask_default, hbw_nodemask_env_variable))

    def test_TC_MEMKIND_hbw_detection_negative_hbw_malloc(self):
        """ This test sets unsupported value of MEMKIND_HBW_NODES,
            then try to perform a successful allocation from DRAM using
            hbw_malloc() thanks to default HBW_POLICY_PREFERRED policy """
        command = "MEMKIND_HBW_NODES=-1 " \
            + self.cmd_helper.get_command_path(self.environ_err_test)
        output, retcode = self.cmd_helper.execute_cmd(command, sudo=False)
        assert retcode != 0, self.fail_msg.format(
            "\nError: Execution of: \'{0}\' returns: {1} \noutput: {2}".format(
                command, retcode, output))
        assert self.expected_libnuma_warning == output, self.fail_msg.format(
            "Error: expected libnuma warning ({0}) "
            "was not found (output: {1})").format(
                self.expected_libnuma_warning, output)



# SPDX-License-Identifier: BSD-2-Clause
# Copyright (C) 2020 - 2021 Intel Corporation.

import argparse
import collections
import fabric
import functools
import os
import paramiko
import pathlib
import psutil
import re
import subprocess
import sys
import time
import typing

TCP_PORT = 10022
VNC_DEFAULT_PORT = 5900
VNC_PORT = 5
GUEST_USERNAME = 'memkinduser'
GUEST_PASS = 'memkindpass'
MEMKIND_GUEST_PATH = '/mnt/memkind/'
MEMKIND_MOUNT_TAG = 'memkind_host'
MEMKIND_QEMU_PREFIX = 'utils/qemu'
MEMKIND_DOCKER_PREFIX = 'utils/docker'
TOPOLOGY_ENV_VAR = 'MEMKIND_TEST_TOPOLOGY'
TOPOLOGY_DIR = 'topology'
# TODO handling fsdax??
# TODO handling scaling the memory in xml

QemuCfg = collections.namedtuple(
    'QemuCfg', ['workdir', 'image', 'interactive', 'hwloc',
                'force_reinstall', 'verbose', 'run_test',
                'topologies', 'codecov'])
TopologyCfg = collections.namedtuple(
    'TopologyCfg', ['name', 'hmat', 'cpu_model', 'cpu_options', 'mem_options'])


def _logger(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        print(f'Start {func.__name__}')
        result = func(*args, **kwargs)
        print(f'Exit {func.__name__}')
        return result
    return wrapper


class MemoryHostException(Exception):
    """
    Exception raised for errors in case of insufficient memory on host.
    """

    def __init__(self, total_memory: int,
                 tpg_memory: int, tpg_name: str) -> None:
        self.total_memory = total_memory
        self.tpg_memory = tpg_memory
        self.tpg = tpg_name

    def __str__(self) -> str:
        return (f'Memory on host) {self.total_memory} bytes are not enough '
                f'for memory required by'
                f' {self.tpg} : {self.tpg_memory} bytes.')


class GuestConnection:

    CONNECT_TIMEOUT = 30

    def __init__(self, cfg: QemuCfg,
                 topology_name: str, qemu_pid: int) -> None:
        self.run_test = cfg.run_test
        self.force_reinstall = cfg.force_reinstall
        self.verbose = cfg.verbose
        self.codecov = cfg.codecov
        self.hwloc = cfg.hwloc
        self.topology_name = topology_name
        self.qemu_pid = qemu_pid

    @property
    def _verbosity_option(self) -> dict:
        """
        Verbosity option
        """
        return {'echo': True} if self.verbose else {'hide': True}

    @property
    def _connection_params(self) -> dict:
        """
        Connection parameters Host->Guest
        """
        return {'host': 'localhost',
                'user': f'{GUEST_USERNAME}',
                'connect_timeout': f'{self.CONNECT_TIMEOUT}',
                'connect_kwargs': {"password": f'{GUEST_PASS}'},
                'port': f'{TCP_PORT}'
                }

    @property
    def _memkind_configure_params(self) -> str:
        """
        Memkind configure parameters
        """
        config_params = '--prefix=/usr'
        if not self.hwloc:
            config_params += ' --disable-hwloc'
        if self.codecov:
            config_params += ' --enable-gcov'
        return config_params

    @_logger
    def _build_and_reinstall_memkind(self, c: fabric.Connection) -> None:
        """
        Rebuild and reinstall memkind library
        """
        result = c.run('nproc', **self._verbosity_option)
        nproc = int(result.stdout.strip())
        c.run('make distclean', **self._verbosity_option, warn=True)
        c.run('./autogen.sh', **self._verbosity_option)
        c.run(f'./configure {self._memkind_configure_params}',
              **self._verbosity_option)
        c.run(f'make -j {nproc}', **self._verbosity_option)
        c.run(f'make checkprogs -j {nproc}', **self._verbosity_option)
        c.run('sudo make uninstall', **self._verbosity_option)
        c.run('sudo make install', **self._verbosity_option)
        c.run('sudo ldconfig', **self._verbosity_option)

    @_logger
    def _mount_workdir(self, c: fabric.Connection) -> None:
        """
        Mount working directory from host to guest
        """
        c.run(f'sudo mkdir -p {MEMKIND_GUEST_PATH}', echo=True)
        c.run(f'sudo mount {MEMKIND_MOUNT_TAG} '
              f'{MEMKIND_GUEST_PATH} -t 9p -o trans=virtio', echo=True)

    def _set_test_env_name(self, c: fabric.Connection) -> None:
        """
        Set environment variable for topology
        """
        if self.topology_name:
            c.config.run.env = {f'{TOPOLOGY_ENV_VAR}': f'{self.topology_name}'}

    @_logger
    def _run_codecov_script(self, c: fabric.Connection) -> None:
        """
        Run code coverage script
        """
        codecov_script = pathlib.Path(
            MEMKIND_GUEST_PATH, MEMKIND_DOCKER_PREFIX,
            'docker_run_coverage.sh')
        c.config.run.env['TEST_SUITE_NAME'] = self.topology_name
        c.config.run.env['CODECOV_TOKEN'] = self.codecov
        c.run(f'{codecov_script} {MEMKIND_GUEST_PATH}',
              **self._verbosity_option)

    @_logger
    def _run_memkind_test(self, c: fabric.Connection) -> None:
        """
        Run memkind test
        """
        c.run('make unit_tests_hmat', echo=True)
        if self.codecov:
            self._run_codecov_script(c)

    @_logger
    def _shutdown(self, c: fabric.Connection) -> None:
        """
        Shutdown guest
        """
        c.run('sudo shutdown now', echo=True, warn=True)
        _, alive = psutil.wait_procs([psutil.Process(
            self.qemu_pid)], timeout=10)
        for p in alive:
            p.kill()

    @_logger
    def verify_connection(self) -> None:
        """
        Verify if connection is available
        """
        for sec in range(self.CONNECT_TIMEOUT):
            try:
                with fabric.Connection(**self._connection_params) as c:
                    c.run('whoami', hide='both')
            except paramiko.SSHException:
                print(f'FAILURE: SSH connection'
                      f' to guest after {sec} sec - retrying.')
                time.sleep(1)
            else:
                print(f'SUCCESS: SSH connection to guest after {sec} sec.')
                return
        sys.exit(f'SSH connection to guest fails after'
                 f' {self.CONNECT_TIMEOUT} sec.')

    @_logger
    def run_connection_test(self, force_reinstall: bool) -> None:
        """
        Run QEMU in test mode
        """
        with fabric.Connection(**self._connection_params,
                               inline_ssh_env=True) as c:
            try:
                self._mount_workdir(c)
                with c.cd(MEMKIND_GUEST_PATH):
                    self._set_test_env_name(c)
                    if force_reinstall:
                        self._build_and_reinstall_memkind(c)
                    if self.run_test:
                        self._run_memkind_test(c)
            finally:
                self._shutdown(c)

    @_logger
    def run_connection_dev(self, force_reinstall: bool) -> None:
        """
        Run QEMU in dev mode
        """
        with fabric.Connection(**self._connection_params,
                               inline_ssh_env=True) as c:
            try:
                self._mount_workdir(c)
                with c.cd(MEMKIND_GUEST_PATH):
                    self._set_test_env_name(c)
                    if force_reinstall:
                        self._build_and_reinstall_memkind(c)
                print(f'QEMU development is ready to use via SSH'
                      f' ({TCP_PORT}) or VNC ({VNC_DEFAULT_PORT + VNC_PORT})')
            except fabric.exceptions.GroupException:
                self._shutdown(c)


class QEMU:

    def __init__(self, cfg: QemuCfg) -> None:
        self.cfg = cfg
        self.pid = None

    @property
    def _qemu_exec(self) -> str:
        """
        QEMU binary
        """
        return 'qemu-system-x86_64'

    @property
    def _hda_option(self) -> str:
        """
        Hard drive as a file
        """
        return f'-hda {self.cfg.image}'

    @property
    def _optim_option(self) -> str:
        """
        Optimization options:
        - Use kernel based virtual machine
        """
        return '-enable-kvm'

    def _cpu_model(self, tpg: TopologyCfg) -> str:
        """
        CPU Model
        """
        # virsh doesn't support model KnightsMill
        if "KnightsMill" in tpg.name:
            cpu_model = 'KnightsMill'
        else:
            cpu_model = 'host' if tpg.cpu_model == 'qemu64' else tpg.cpu_model
        return f'-cpu {cpu_model}'

    @property
    def _connect_option(self) -> str:
        """
        Connect options:
        - VNC on port 5
        - tcp on port 10022
        """
        return (f'-vnc :{VNC_PORT} -netdev user,'
                f'id=net0,hostfwd=tcp::{TCP_PORT}-:22 '
                f'-device virtio-net,netdev=net0')

    @property
    def _boot_option(self) -> str:
        """
        Boot order
        - Start from first virtual hard drive
        """
        return '-boot c'

    def _cpu_options(self, tpg: TopologyCfg) -> str:
        """
        CPU used by QEMU
        """
        return tpg.cpu_options if tpg.cpu_options else '-smp 2'

    def _memory_option(self, tpg: TopologyCfg) -> str:
        """
        Memory used by QEMU
        """
        if tpg.mem_options:
            mem_needed = 0
            all_sizes = re.findall(r'size=(\d+)', tpg.mem_options)
            for single_size in all_sizes:
                mem_needed += int(single_size)

            mem = psutil.virtual_memory()
            if mem_needed >= mem.total:
                raise MemoryHostException(mem.total, mem_needed, tpg.name)
            return f'-m {mem_needed/1024/1024}M'
        else:
            return '-m 2G'

    @property
    def _daemonize_option(self) -> str:
        """
        Daemonize option
        - run process as daemon or in terminal
        """
        return '-nographic' if self.cfg.interactive else '-daemonize'

    def _hmat_options(self, tpg: TopologyCfg) -> str:
        """
        HMAT option
        """
        return '-machine hmat=on' if tpg.hmat else ''

    def _memory_topology(self, tpg: TopologyCfg) -> str:
        """
        Memory topology string
        """
        mem_tpg = tpg.mem_options
        # workaround for QEMU - validation in machine_set_cpu_numa_node method
        if tpg.hmat:
            mem_tpg_parsed = re.sub(r'cpus=(\d+\-\d+,)', '', tpg.mem_options)
            mem_tpg_list = list(mem_tpg_parsed.partition('-numa dist'))
            for numa_id in re.findall(r'nodeid=(\d+),cpus=', tpg.mem_options):
                cpu_tpg = f'-numa cpu,node-id={numa_id},socket-id={numa_id} '
                mem_tpg_list.insert(1, cpu_tpg)
            mem_tpg = ''.join(mem_tpg_list)
        return mem_tpg

    @property
    def _mount_option(self) -> str:
        """
        mount option
        - Start from first virtual hard drive
        """
        return (f'-fsdev local,security_model=passthrough,'
                f'id=fsdev0,path={self.cfg.workdir} '
                f'-device virtio-9p-pci,id=fs0,fsdev=fsdev0,'
                f'mount_tag={MEMKIND_MOUNT_TAG}')

    def _qemu_cmd(self, tpg: TopologyCfg) -> typing.List[str]:
        """
        QEMU command concat
        """
        cmd_str = ' '.join([self._qemu_exec,
                            self._hda_option,
                            self._optim_option,
                            self._cpu_model(tpg),
                            self._connect_option,
                            self._boot_option,
                            self._memory_option(tpg),
                            self._memory_topology(tpg),
                            self._cpu_options(tpg),
                            self._hmat_options(tpg),
                            self._mount_option,
                            self._daemonize_option])
        if self.cfg.verbose:
            print(f'QEMU command: {cmd_str}')
        return cmd_str.split()

    def _start_qemu_proc(self, qemu_cmd: typing.List[str]) -> None:
        """
        Start QEMU process
        """
        try:
            result = subprocess.run(qemu_cmd, stderr=subprocess.PIPE)
            result.check_returncode()
            for proc in psutil.process_iter():
                if proc.name() == self._qemu_exec:
                    self.pid = proc.pid
        except subprocess.CalledProcessError:
            sys.exit(f'\nQEMU cannot start process:'
                     f' image {self.cfg.image}: {result.stderr}.')

    def run(self) -> None:
        """
        run QEMU
        """
        reinstall_opt = self.cfg.force_reinstall
        for topology in self.cfg.topologies:
            try:
                qemu_cmd = self._qemu_cmd(topology)
            except MemoryHostException as e:
                print(f' QEMU ERROR: Memory topology skipped {topology.name}')
                print(f'{e}')
                continue
            if self.cfg.interactive:
                for sec in reversed(range(5)):
                    print(f'QEMU will start in {sec} seconds.')
                    time.sleep(1)
                self._start_qemu_proc(qemu_cmd)
            else:
                self._start_qemu_proc(qemu_cmd)
                con = GuestConnection(self.cfg, topology.name, self.pid)
                con.verify_connection()
                if self.cfg.run_test:
                    con.run_connection_test(reinstall_opt)
                else:
                    con.run_connection_dev(reinstall_opt)
                reinstall_opt = False


def find_root_memkind_dir() -> str:
    """
    Find memkind root directory on host
    """
    try:
        result = subprocess.run(['git', 'rev-parse', '--show-toplevel'],
                                stdout=subprocess.PIPE)
        result.check_returncode()
        path = result.stdout.decode('utf-8').strip()
    except subprocess.CalledProcessError:
        # Get root directory without git modify
        # this if we change the path of script
        path = str(pathlib.Path(__file__).parents[2])
    return path


def parse_topology_xml(tpg_file_name: str) -> TopologyCfg:
    """
    Parse topology xml file
    """
    try:
        result = subprocess.run(['virsh', 'domxml-to-native', 'qemu-argv',
                                tpg_file_name], stdout=subprocess.PIPE)
        result.check_returncode()
        libvirt_args = result.stdout.decode('utf-8').strip()
        tpg_cfg = {'name': re.search(r'guest=(\w+)', libvirt_args).group(1),
                   'hmat': 'hmat=on' in libvirt_args,
                   'cpu_model': re.search(r'cpu (\S+)', libvirt_args).group(1),
                   'cpu_options': re.search(
                       '(?=-smp)(.*)threads=[0-9]+', libvirt_args).group(0),
                   'mem_options': re.search(
                       '(?=-object memory)(.*)(?=-uuid)',
                       libvirt_args).group(1)}

        tpg = TopologyCfg(**tpg_cfg)
    except FileNotFoundError:
        sys.exit(f'\n XML file: {tpg_file_name} is missing.')
    except subprocess.CalledProcessError:
        sys.exit(f'\n XML file: {tpg_file_name} error in virsh parsing')
    return tpg


def parse_topology(cfg: dict) -> typing.List[TopologyCfg]:
    """
    Parse memory topology XML file
    """
    tpg_list = []
    single_topology = cfg.get('topology', None)
    if single_topology:
        tpg_list.append(parse_topology_xml(single_topology))
    elif cfg['mode'] == 'test':
        topology_dir = pathlib.Path(
            find_root_memkind_dir(), MEMKIND_QEMU_PREFIX, TOPOLOGY_DIR)
        for filepath in pathlib.Path.iterdir(topology_dir):
            if filepath.suffix == '.xml':
                tpg_list.append(parse_topology_xml(str(filepath)))
    else:
        dev_topology = {'name': '', 'hmat': False,
                        'cpu_options': '', 'mem_options': ''}
        tpg_list.append(TopologyCfg(**dev_topology))
    return tpg_list


def validate_image_path(image_path: str) -> str:
    """
    Validate QEMU image file path
    """
    if not os.path.exists(image_path):
        sys.exit(f"\n Image file: {image_path} doesn't exist")
    return image_path


def parse_arguments() -> QemuCfg:
    """
    Parse command line arguments
    """
    qemu_cfg = {'workdir': find_root_memkind_dir()}
    parser = argparse.ArgumentParser()
    parser.add_argument('--image', help='QEMU image', required=True)
    parser.add_argument('--mode', choices=['dev', 'test'],
                        help='execution mode', default='dev')
    parser.add_argument('--hwloc', choices=['on', 'off'],
                        help='hwloc support', default='on')
    parser.add_argument('--topology', help='memory topology XML file')
    parser.add_argument('--codecov', help='upload token for Codecov')
    parser.add_argument('--interactive', action="store_true",
                        help='execute an interactive bash shell',
                        default=False)
    parser.add_argument('--force_reinstall', action="store_true",
                        help='force rebuild and install memkind',
                        default=False)
    parser.add_argument('--verbose', action="store_true",
                        help='run script in verbose mode', default=False)
    cli_cfg = vars(parser.parse_args())

    if cli_cfg['interactive'] and cli_cfg['mode'] == 'test':
        parser.error("interactive is supported only by dev mode")

    if cli_cfg['codecov'] and cli_cfg['mode'] == 'dev':
        parser.error("codecov is supported only by test mode")

    qemu_cfg['image'] = validate_image_path(cli_cfg['image'])
    qemu_cfg['interactive'] = cli_cfg['interactive']
    qemu_cfg['force_reinstall'] = cli_cfg['force_reinstall']
    qemu_cfg['verbose'] = cli_cfg['verbose']
    qemu_cfg['topologies'] = parse_topology(cli_cfg)
    qemu_cfg['run_test'] = cli_cfg['mode'] == 'test'
    qemu_cfg['hwloc'] = cli_cfg['hwloc'] == 'on'
    qemu_cfg['codecov'] = cli_cfg['codecov']
    return QemuCfg(**qemu_cfg)


@_logger
def main() -> None:
    """
    Main function
    """
    qemu_cfg = parse_arguments()
    qemu = QEMU(qemu_cfg)
    qemu.run()


if __name__ == "__main__":
    main()
